"""Configuration system for embedded mechanical."""
import os


class Configuration:
    """Configuration class for Mechanical."""

    def __init__(self):
        """Construct a new Configuration instance."""
        self._no_act_addins = False
        self._no_wb1_addins = False

    @property
    def no_act_addins(self) -> bool:
        """Property to disable all ACT Addins."""
        return self._no_act_addins

    @no_act_addins.setter
    def no_act_addins(self, value: bool):
        """Setter of property to disable all ACT Addins."""
        self._no_act_addins = value

    @property
    def no_wb1_addins(self) -> bool:
        """Property to disable all WB1 Addins."""
        return self._no_wb1_addins

    @no_wb1_addins.setter
    def no_wb1_addins(self, value: bool):
        """Setter of property to disable all WB1 Addins."""
        self._no_wb1_addins = value


def configure(configuration: Configuration):
    """Apply the given configuration."""
    if configuration.no_act_addins:
        os.environ["ANSYS_MECHANICAL_STANDALONE_NO_ACT_EXTENSIONS"] = "1"

    # TODO - support a configuration option for no wb1 addins

    # TODO - support fine-grained configuration options, design an interface for WB1
    #        to accept these and have the ability to call that interface before
    #        loading WB. (perhaps WB.Initialize can take some new parameters!)



"""Main application class for embedded Mechanical."""
import atexit
import os
import typing

from ansys.mechanical.core.embedding import initializer, runtime
from ansys.mechanical.core.embedding.config import Configuration, configure


def _get_available_versions() -> typing.Dict[int, str]:
    supported_versions = [222, 231, 232]
    if os.name == "nt":  # pragma: no cover
        supported_versions = [222, 231, 232]
        awp_roots = {ver: os.environ.get(f"AWP_ROOT{ver}", "") for ver in supported_versions}
        installed_versions = {
            ver: path for ver, path in awp_roots.items() if path and os.path.isdir(path)
        }
        return installed_versions
    else:
        # TODO - this assumes the env var is set.
        return {232: os.environ["AWP_ROOT232"]}


def _get_default_version() -> int:
    vers = _get_available_versions()
    if not vers:  # pragma: no cover
        raise Exception(
            "Appropriate version of Ansys Mechanical is not installed. Must be at least v222"
        )
    return max(vers.keys())


def _get_default_configuration() -> Configuration:
    configuration = Configuration()
    if os.name != "nt":
        configuration.no_act_addins = True
    return configuration


INSTANCES = []


def _dispose_embedded_app(instances):  # pragma: nocover
    if len(instances) > 0:
        instance = instances[0]
        instance._dispose()


class App:
    """Mechanical embedding Application."""

    def __init__(self, db_file=None, **kwargs):
        """Construct an instance of the mechanical Application.

        db_file is an optional path to a mechanical database file (.mechdat or .mechdb)
        you may set a version number with the `version` keyword argument.
        """
        global INSTANCES
        if len(INSTANCES) > 0:
            raise Exception("Cannot have more than one embedded mechanical instance")
        self._version = kwargs.get("version")
        if self._version == None:
            self._version = _get_default_version()
        initializer.initialize(self._version)
        import clr

        clr.AddReference("Ansys.Mechanical.Embedding")
        import Ansys

        configuration = kwargs.get("config", _get_default_configuration())
        configure(configuration)
        self._app = Ansys.Mechanical.Embedding.Application(db_file)
        runtime.initialize(self._version)
        self._disposed = False
        if len(INSTANCES) == 0:
            atexit.register(_dispose_embedded_app, INSTANCES)
        INSTANCES.append(self)

    def __repr__(self):
        """Get the product info."""
        if self._version < 232:  # pragma: no cover
            return "Ansys Mechanical"
        import clr

        clr.AddReference("Ansys.Mechanical.Application")
        import Ansys

        return Ansys.Mechanical.Application.ProductInfo.ProductInfoAsString

    def __enter__(self):  # pragma: no cover
        """Enter the scope."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # pragma: no cover
        """Exit the scope."""
        self._dispose()

    def _dispose(self):
        if self._disposed:
            return
        self._app.Dispose()
        self._disposed = True

    def open(self, db_file):
        """Open the db file."""
        self.DataModel.Project.Open(db_file)

    def save(self, path=None):
        """Save the project."""
        self.DataModel.Project.Save(path)

    def save_as(self, path):
        """Save the project as."""
        self.DataModel.Project.SaveAs(path)

    def new(self):
        """Clear to a new application."""
        self.DataModel.Project.New()

    def execute_script(self, script: str):
        """Execute the given script with the internal IronPython engine."""
        SCRIPT_SCOPE = "pymechanical-internal"
        if not hasattr(self, "script_engine"):
            import clr

            clr.AddReference("Ansys.Mechanical.Scripting")
            import Ansys

            engine_type = Ansys.Mechanical.Scripting.ScriptEngineType.IronPython
            script_engine = Ansys.Mechanical.Scripting.EngineFactory.CreateEngine(engine_type)
            empty_scope = False
            debug_mode = False
            script_engine.CreateScope(SCRIPT_SCOPE, empty_scope, debug_mode)
            self.script_engine = script_engine
        light_mode = True
        args = None
        rets = None
        return self.script_engine.ExecuteCode(script, SCRIPT_SCOPE, light_mode, args, rets)

    @property
    def DataModel(self):
        """Return the DataModel."""
        return self._app.DataModel

    @property
    def ExtAPI(self):
        """Return the ExtAPI object."""
        return self._app.ExtAPI

    @property
    def version(self):
        """Returns the version of the app."""
        return self._version



"""Initialize the package level imports."""
import logging
import os

import appdirs

from ansys.mechanical.core.logging import Logger

# Create logger for package level use
LOG = Logger(level=logging.ERROR, to_file=False, to_stdout=True)
LOG.debug("Loaded logging module as LOG")

from ansys.mechanical.core._version import __version__
from ansys.mechanical.core.mechanical import (
    change_default_mechanical_path,
    close_all_local_instances,
    find_mechanical,
    get_mechanical_path,
    launch_mechanical,
)

# import few classes / functions
from ansys.mechanical.core.mechanical import Mechanical as Mechanical

try:
    from ansys.mechanical.core.embedding import App, global_variables

    HAS_EMBEDDING = True
except:
    HAS_EMBEDDING = False

# manage the package level ports
LOCAL_PORTS = []

from ansys.mechanical.core.pool import LocalMechanicalPool

BUILDING_GALLERY = False

# Setup data directory
try:
    USER_DATA_PATH = appdirs.user_data_dir("ansys_mechanical_core")
    if not os.path.exists(USER_DATA_PATH):
        os.makedirs(USER_DATA_PATH)

    EXAMPLES_PATH = os.path.join(USER_DATA_PATH, "examples")
    if not os.path.exists(EXAMPLES_PATH):
        os.makedirs(EXAMPLES_PATH)

except Exception:
    pass



"""Logging module, used to setup logging options for embedded mechanical."""
from pathlib import Path


def _get_embedding():
    import clr

    try:
        clr.AddReference("Ansys.Mechanical.Embedding")
        import Ansys

        return Ansys.Mechanical.Embedding
    except:
        raise Exception("Cannot use logging until after mechanical embedding is initialized")


class Logging:
    """Logging class for embedded Mechanical."""

    class Level:
        """Logging level for Mechanical."""

        TRACE = 0
        DEBUG = 1
        INFO = 2
        WARNING = 3
        ERROR = 4
        FATAL = 5

    @classmethod
    def config(cls, **kwargs):
        """Similar to logging.basicConfig.

        Available options:
        filename -> str
        level -> Logging.Level values
        auto_flush -> bool
        enabled -> bool
        stdout -> bool
        """
        config = _get_embedding().LoggingConfiguration
        if "filename" in kwargs:
            path = Path(kwargs["filename"])
            config.Filename = path.name
            config.Path = str(path.parent)
        if "enabled" in kwargs:
            config.Enabled = kwargs["enabled"]
        if "auto_flush" in kwargs:
            config.AutoFlush = kwargs["auto_flush"]
        if "stdout" in kwargs:
            config.LogToStdOut = kwargs["stdout"]
        if "level" in kwargs:
            config.FilterLevel = kwargs["level"]

    @classmethod
    def log_message(cls, severity: Level, context: str, message: str) -> None:
        """Log the message to the configured logging mechanism."""
        _get_embedding().Logging.LogMessage(severity, context, message)

    @classmethod
    def can_log_message(cls, severity: Level) -> bool:
        """Return whether a message with the given severity will be output into the log."""
        return _get_embedding().Logging.CanLogMessage(severity)

    @classmethod
    def flush(cls) -> None:
        """Flushes the log manually."""
        return _get_embedding().Logging.Flush()



"""This is the .NET assembly resolving for embedding Ansys Mechanical.

Note that for some Mechanical Addons - additional resolving may be
necessary. A resolve handler is shipped with Ansys Mechanical on windows
starting in version 23.1 and on linux starting in version 23.2
"""


def resolve(version):
    """Resolve function for all versions of Ansys Mechanical."""
    import clr  # isort: skip
    import System  # isort: skip

    clr.AddReference("Ansys.Mechanical.Embedding")
    import Ansys  # isort: skip

    assembly_resolver = Ansys.Mechanical.Embedding.AssemblyResolver
    if version == 231:  # pragma: no cover
        resolve_handler = assembly_resolver.WindowsResolveEventHandler
    else:
        resolve_handler = assembly_resolver.MechanicalResolveEventHandler
    System.AppDomain.CurrentDomain.AssemblyResolve += resolve_handler



"""Initializer for Mechanical embedding. Sets up paths and resolvers."""
import os
from pathlib import Path
import sys

from ansys.mechanical.core.embedding.loader import load_clr
from ansys.mechanical.core.embedding.resolver import resolve

INITIALIZED_VERSION = None


def __add_sys_path(version) -> str:
    install_path = Path(os.environ[f"AWP_ROOT{version}"])
    platform_string = "winx64" if os.name == "nt" else "linx64"
    bin_path = install_path / "aisol" / "bin" / platform_string
    sys.path.append(str(bin_path.resolve()))


def __disable_sec() -> None:
    """SEC is part of RSM and is unstable with embedding.

    I'm not going to debug why that is since we are planning to support
    DCS/REP in the future instead of RSM.
    """
    os.environ["ANSYS_MECHANICAL_EMBEDDING_NO_SEC"] = "1"


def initialize(version):
    """Initialize Mechanical embedding."""
    global INITIALIZED_VERSION
    if INITIALIZED_VERSION != None:
        assert INITIALIZED_VERSION == version
        return
    INITIALIZED_VERSION = version

    __disable_sec()

    # need to add system path in order to import the assembly with the resolver
    __add_sys_path(version)

    # load the CLR with mono that is shipped with the unified ansys installer
    load_clr(os.environ[f"AWP_ROOT{version}"])

    # attach the resolver
    resolve(version)



"""Shims for embedded Mechanical.

These shims are used when APIs are released in newer versions of Mechanical,
but workarounds exist in an older release
"""


def import_materials(
    app: "ansys.mechanical.core.embedding.Application", material_file: str
) -> None:
    """Import material from matml file."""
    if app._version >= 232:
        materials = app.DataModel.Project.Model.Materials
        materials.Import(material_file)
    else:  # pragma: no cover
        material_file = material_file.replace("\\", "\\\\")
        script = 'DS.Tree.Projects.Item(1).LoadEngrDataLibraryFromFile("' + material_file + '");'
        app.ExtAPI.Application.ScriptByName("jscript").ExecuteCommand(script)



"""Additional imports for embedded Mechanical."""
import typing


def global_variables(app: "ansys.mechanical.core.App") -> typing.Dict:
    """Return the Mechanical scripting global variables as a dict.

    It can be used to add all of these as global variables in python
    with this command:
        `globals().update(global_variables(embedded_app))`
    """
    vars = {}
    vars["ExtAPI"] = app.ExtAPI
    vars["DataModel"] = app.DataModel
    vars["Model"] = app.DataModel.Project.Model
    vars["Tree"] = app.DataModel.Tree
    import clr  # isort: skip

    clr.AddReference("System.Collections")
    from Ansys.Core.Units import Quantity

    import System  # isort: skip
    import Ansys  # isort: skip

    vars["Quantity"] = Quantity
    vars["System"] = System
    vars["Ansys"] = Ansys
    return vars



"""Runtime initialize for pythonnet in embedding."""
from ansys.mechanical.core.embedding import loader


def __initialize_runtime_pythonnet_3():
    """Add the codecs that are in python3.

    These are needed for many things, including
    Mechanical's list conversions (python list to C# lists).
    """
    import Python.Runtime.Codecs as codecs

    codecs.ListDecoder.Instance.Register()
    codecs.SequenceDecoder.Instance.Register()
    codecs.IterableDecoder.Instance.Register()
    # TODO - FunctionCodec


def initialize(version: int) -> None:
    """Initialize the runtime.

    Pythonnet is already initialized but we need to
    do some special codec handling to make sure the
    interop works well for Mechanical.
    """
    if loader.is_pythonnet_3():
        __initialize_runtime_pythonnet_3()
    else:  # pragma: no cover
        # pythonnet 2.5 is supported with some codecs
        # and some additions to the system path that are shipped with Mechanical in 2023 R1
        # these additions to the system path may not be desirable for pymechanical embedding
        if version == 231:
            import clr

            clr.AddReference("Ansys.Mechanical.CPython")
            import Ansys

            Ansys.Mechanical.CPython.CPythonEngine.InitializeRuntime()



"""clr_loader for pymechanical embedding. This loads the CLR on both windows and linux."""
import os

try:
    from importlib.metadata import version

    HAS_IMPORTLIB = True
except:  # pragma: no cover
    # TODO - only support importlib.metadata::version after dropping python3.7 support.
    # pkg_resources is part of distutils and is considered obsolete.
    import pkg_resources

    HAS_IMPORTLIB = False


def __get_clr_loader_version():
    if HAS_IMPORTLIB:
        return version("clr_loader")
    else:  # pragma: no cover
        return pkg_resources.get_distribution("clr_loader").version


def __get_mono(assembly_dir, config_dir):
    import clr_loader

    if __get_clr_loader_version() == "0.2.5":
        libmono = os.path.join(assembly_dir, "libmonosgen-2.0.so")
        mono = clr_loader.get_mono(
            set_signal_chaining=True,
            libmono=libmono,
            assembly_dir=assembly_dir.encode("utf-8"),
            config_dir=config_dir.encode("utf-8"),
        )
    else:
        # the bugs with get_mono are fixed (clr_loader PR #48)
        mono = clr_loader.get_mono(
            set_signal_chaining=True, assembly_dir=assembly_dir, config_dir=config_dir
        )
    return mono


def load_clr_mono(install_loc):
    """Load the clr using mono that is shipped with the unified install."""
    from pythonnet import load

    mono_dir = os.path.join(install_loc, "Tools", "mono", "Linux64")
    assembly_dir = os.path.join(mono_dir, "lib")
    config_dir = os.path.join(mono_dir, "etc")
    mono = __get_mono(assembly_dir, config_dir)
    load(mono)


def load_clr(install_loc: str) -> None:
    """Load the clr, the outcome of this function is that `clr` is usable."""
    if os.name == "nt":  # pragma: no cover
        return
    load_clr_mono(install_loc)


def is_pythonnet_3() -> bool:
    """Return whether pythonnet version 3 is used."""
    import clr

    return 3 == int(clr.__version__.split(".")[0])
